from datetime import datetime
from math import isclose
import pandas as pd
import re


def create_athena_query(queries):

    final_queries = []

    for query in queries:
        if 'query' in query:
            final_queries.append({'query_name' : query['query_name'],
                        'query' : query['query'],
                        })
            final_queries[-1]['values_to_check'] = get_group_by_length(final_queries[-1]['query'])
            continue
        if 'database' not in query:
            raise Exception(f'missing "database" in query: {query}')
        if 'query_name' not in query:
            raise Exception(f'missing "query_name" in query: {query}')
        if 'select' not in query:
            raise Exception(f'missing "method" in query: {query}')

        final_queries.append({'query_name' : query['query_name'],
                        'query' : create_one_query(query['database'],
                                                    query['select'],
                                                    query['condition'],
                                                    query['groupBy'])
                        })
        final_queries[-1]['values_to_check'] = get_group_by_length(final_queries[-1]['query'])
    return final_queries

def get_group_by_length(query):
    query = re.sub(' +', ' ', query.lower()) #Remove unwanted spaces
    if 'group by' in query:
        num_of_select = len(query.split('from')[0].split(','))
        num_of_group_by = len(query.split('group by')[-1].split(','))
        return num_of_select - num_of_group_by
    else:
        return 1
    

def create_one_query(db, select, condition, group_by):
    final_query = f'SELECT {", ".join(select)}\
            FROM {db}'
    if condition:
        final_query += f' WHERE {" AND ".join(condition)}'
    if group_by:
        final_query += f' GROUP BY {group_by}'

    return final_query

def create_test_result(tests, queries_result):
    final_result= {}
    for test_idx, test in enumerate(tests):
        rel_tol = test.get('rel_tol', 0.01)
        field_1 = handle_field(test['field_1'], queries_result)
        field_2 = handle_field(test['field_2'], queries_result)
        test_name = f'test_{test_idx}'
        values_to_check = field_1.get('values_to_check', False) or field_2.get('values_to_check', False) 
        final_result[test_name] = get_test_result(field_1['result'], field_2['result'],
                                                 test['method'],values_to_check, rel_tol)

    date = str(datetime.now().date())
    time_stamp = int(datetime.now().timestamp())
    file_name= f'{time_stamp}.csv'
    s3_path = f's3://bigabid-test/athena_tests/{date}/{file_name}'
    dict_to_df = {(i,j): final_result[i][j]
                           for i in final_result.keys() 
                           for j in final_result[i].keys()}

    result_to_df=pd.DataFrame.from_dict(dict_to_df, orient='index')
    print(result_to_df)
    result_to_df['test_date'] = date
    result_to_df['time_stamp'] = time_stamp
    result_to_df.to_csv(s3_path, sep='\t', header=False)
    print(f'result has been saved to {s3_path}')
    print(f'result can be queried as the following: \
    SELECT * \
    FROM temp_tables.athena_tests\
    WHERE time_stamp={time_stamp}')

def handle_field(field, queries):
    if field['type'] == 'query':
        return queries[field['value']]
    return {'result' : field['value']}

def get_test_result(field_1, field_2, method, values_to_check, rel_tol):
    test_final_result = dict()
    sort_by_length = len(field_1[0]) - values_to_check
    sorted_field_1 = sorted(field_1, key=lambda x:x[:sort_by_length]) 
    sorted_field_2 = sorted(field_2, key=lambda x:x[:sort_by_length])
    checking_idx = 0-values_to_check
    for res1, res2 in zip(sorted_field_1, sorted_field_2):
        value_list_1, value_list_2 = res1[checking_idx:], res2[checking_idx:]

        # result are strings, lets try to convert them to floats
        value_list_1 = convert_to_float(value_list_1)
        value_list_2 = convert_to_float(value_list_2) 
        comparing_res = compare_results(value_list_1,value_list_2,method,rel_tol)
        test_final_result[f'{res1} {method} {res2}'] = {'expected' : value_list_1,
                                                    'actual' : value_list_2, 
                                                    'reuslt': comparing_res,
                                                    'properties': f'rel_tol={rel_tol}' if method=='eq' else None}

    
    return test_final_result

def convert_to_float(value_list):
    new_list = []
    for value in value_list:
        try:
            new_list.append(float(value))
        except Exception as e:
            new_list.append(value)
    return new_list

def compare_results(value_list_1, value_list_2, method, rel_tol):
    print(rel_tol)
    if method == 'eq':
        for val_1, val_2 in zip(value_list_1,value_list_2):
            if not isclose(val_1, val_2, rel_tol=rel_tol):
                return False
    elif method == 'gt':
        for val_1, val_2 in zip(value_list_1,value_list_2):
            if not val_1 > val_2:
                return False
    elif method == 'gte':
        for val_1, val_2 in zip(value_list_1,value_list_2):
            if not val_1 >= val_2:
                return False
    elif method == 'lt':
        for val_1, val_2 in zip(value_list_1,value_list_2):
            if not val_1 < val_2:
                return False
    elif method == 'lte':
        for val_1, val_2 in zip(value_list_1,value_list_2):
            if not val_1 <= val_2:
                return False
    return True