from airflow.models import DAG
from airflow.providers.slack.operators.slack import SlackAPIPostOperator
from airflow.models import Variable
from typing import Dict, Optional
import os
 
class BigaFlow():

    __text = """
    {emoji} Task {status}. 
    *Task*: {task}  
    *Dag*: {dag} 
    *Execution Time*: {exec_date}  
    *Log Url*: {log_url} 
    {alerts_to}            
    """

    __biga_args = {
        'catchup': False,
    }

    __biga_default_args = {}
    __environment_run = os.environ.get('AIRFLOW__CORE__ENVIRONMENT', 'development')
    __active_alerts = __environment_run.lower() in ['test', 'production']
    __is_production = __environment_run.lower() == 'production'

    def build_bigadag(self,
              default_args: Optional[Dict]= None, 
              disable_callback=['on_success_callback', 'on_retry_callback'],
              on_failure_callback=None, 
              on_success_callback=None, 
              on_retry_callback=None,
              slack_channel='#airflow_dags_alerts', 
              mentions=None,
              *args, 
              **kwargs):
        self.disable_callback = disable_callback
        self.mentions = mentions
        self.on_failure_callback=on_failure_callback 
        self.on_success_callback=on_success_callback
        self.on_retry_callback=on_retry_callback
        self.slack_channel = slack_channel
        self.default_args=self.__override_default_args(default_args)
        self.default_args=self.__set_tasks_events_callbacks(self.default_args) 
        self.args=args
        self.kwargs=self.__override_kwargs(kwargs)
        self.kwargs.update({
            'default_args': default_args
        }) 
        return DAG(*self.args, **self.kwargs)       
 
    def __return_order_list_of_user(self, alerts_to, mentions):
        user_list = []
        if mentions is not None and len(mentions) > 0:
            for user_name in mentions:
                if (user_name or "").lower() == 'channel':
                    user = "<!channel>"
                else:
                    user = alerts_to[user_name]
                    user = user.replace("@", '')
                    user = f"<@{user}>"
                user_list.append(user)
            user_list = ' ,'.join(user_list)
        return user_list

    def __get_slack_api_operator(self, context, status, slack_channel, emoji, mentions):
        var_name = 'slack_users'
        dag_list = Variable.get(key=var_name, deserialize_json=True)
        alerts_to = self.__return_order_list_of_user(dag_list, mentions)
        slack_channel = slack_channel
        alert = SlackAPIPostOperator(
            slack_conn_id='slack_default',
            task_id=context.get('task_instance').task_id,
            channel=slack_channel,
            text=self.__text.format(
                emoji=emoji,
                status=status,
                task=context.get('task_instance').task_id,
                dag=context.get('task_instance').dag_id,
                exec_date=context.get('execution_date'),
                log_url=context.get('task_instance').log_url,
                alerts_to=alerts_to
            )
        )
        return alert

    def send_slack_message(self, context, status, emoji, slack_channel=None, mentions=None):
        slack_channel = slack_channel or self.slack_channel
        mentions = mentions or self.mentions
        failed_alert = self.__get_slack_api_operator(context, status, slack_channel, emoji, mentions)
        return failed_alert.execute(context=context)

    def __pipeline_failure_callback(self, context):
        if callable(self.on_failure_callback):
            self.on_failure_callback(context)
        status = 'Failed'
        mentions = self.mentions
        slack_channel = self.slack_channel
        emoji = ':red_circle:'
        failed_alert = self.__get_slack_api_operator(
            context, status, slack_channel, emoji, mentions)
        return failed_alert.execute(context=context)

    def __pipeline_success_callback(self, context):
        if callable(self.on_success_callback):
            self.on_success_callback(context)

        status = 'Successed'
        mentions = self.mentions
        slack_channel = self.slack_channel
        emoji = ':white_check_mark:'
        success_alert = self.__get_slack_api_operator(
            context, status, slack_channel, emoji, mentions)
        return success_alert.execute(context=context)

    def __pipeline_retry_callback(self, context):
        if callable(self.on_retry_callback):
            self.on_retry_callback(context)

        status = 'Retry'
        mentions = self.mentions
        slack_channel = self.slack_channel
        emoji = ':repeat_one:'
        retry_alert = self.__get_slack_api_operator(
            context, status, slack_channel, emoji, mentions)
        return retry_alert.execute(context=context)

    def __set_tasks_events_callbacks(self, default_args):
        if self.__active_alerts:
            biga_callbacks = {
                'on_failure_callback': self.__pipeline_failure_callback,
                'on_retry_callback': self.__pipeline_retry_callback,
                'on_success_callback': self.__pipeline_success_callback
            }
            alert_status = []
            if self.on_failure_callback is None and 'on_failure_callback' in self.disable_callback:
                alert_status.append('on_failure_callback')
            if self.on_success_callback is None and 'on_success_callback' in self.disable_callback:
                alert_status.append('on_success_callback')
            if self.on_retry_callback is None and 'on_retry_callback' in self.disable_callback:
               alert_status.append('on_retry_callback')
            for key in alert_status:
                biga_callbacks.pop(key)
            default_args.update(biga_callbacks)
        return default_args   

    def __override_kwargs(self, kwargs):
        if not self.__is_production:            
            kwargs.update(self.__biga_args)
        return kwargs
    
    def __override_default_args(self, default_args):
        if not self.__is_production:                        
            default_args.update(self.__biga_default_args)
        return default_args

bigaFlow = BigaFlow()

BigaDAG = bigaFlow.build_bigadag