from airflow.models.baseoperator import BaseOperator
from airflow.utils.decorators import apply_defaults
from airflow.contrib.hooks.aws_athena_hook import AWSAthenaHook
from datetime import datetime, timedelta
from uuid import uuid4


class AddPartitionAthenaOperator(BaseOperator):

    @apply_defaults
    def __init__(
            self,
            table_name: str,
            database: str,
            location: str,
            timedelta_days: int = 0,
            aws_conn_id=None,
            *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.table_name = table_name
        self.database = database
        self.location = location
        self.timedelta_days = timedelta_days
        self.aws_conn_id = aws_conn_id
        self.hook = self.get_hook()

    def execute(self, context):
        self.hook = self.get_hook()
        execution_date = context.get("ds")
        query = self.generate_query(execution_date)
        self.log.info(query)
        query_execution_id = self.hook.run_query(query,
                                                 query_context={
                                                     'Database': self.database
                                                 },
                                                 result_configuration={
                                                     'OutputLocation': 's3://bigabid-tempfiles/airflow-logs/athena/'
                                                 },
                                                 client_request_token=str(uuid4()))

        query_status = self.hook.poll_query_status(query_execution_id, max_tries=5)

        if query_status in AWSAthenaHook.FAILURE_STATES:
            error_message = self.hook.get_state_change_reason(query_execution_id)
            raise Exception(f'Final state of Athena job is {query_status}, query_execution_id is {query_execution_id}. Error: {error_message}')
        elif not query_status or query_status in AWSAthenaHook.INTERMEDIATE_STATES:
            raise Exception(
                f'Final state of Athena job is {query_status}. '
                f'Max tries of poll status exceeded, query_execution_id is {query_execution_id}.')

        results_dict = self.hook.get_query_results(query_execution_id)
        if results_dict['ResponseMetadata']['HTTPStatusCode'] == 200:
            return query_execution_id
        else:
            raise ValueError(f"Error Adding Partition to {self.table_name}")

    def get_hook(self):
        return AWSAthenaHook(self.aws_conn_id)

    def generate_query(self, execution_date):
        date = datetime.fromisoformat(execution_date) - timedelta(days=self.timedelta_days)
        year = date.strftime("%Y")
        month = date.strftime("%m")
        day = date.strftime("%d")

        location = self.location.format(year=year, month=month, day=day)
        return f'ALTER TABLE {self.table_name} ADD PARTITION (YEAR = {year}, MONTH = {month}, DAY = {day}) LOCATION \'{location}\''
