from airflow.models.baseoperator import BaseOperator
from airflow.utils.decorators import apply_defaults
from airflow.contrib.hooks.aws_athena_hook import AWSAthenaHook
from cached_property import cached_property
from airflow.models import Variable
import json
import inspect

from sqlalchemy.sql.functions import Function

class AthenaResultOperator(BaseOperator):
    """
    Asks for the state of the Query until it reaches a failure state or success state.
    If the query fails, the task will fail.
    :param query_execution_id: query_execution_id to check the state of
    :type query_execution_id: str
    :param max_retries: Number of times to poll for query state before
        returning the current state, defaults to None
    :type max_retries: int
    :param aws_conn_id: aws connection to use, defaults to 'aws_default'
    :type aws_conn_id: str
    :param sleep_time: Time in seconds to wait between two consecutive call to
        check query status on athena, defaults to 10
    :type sleep_time: int
    """
    template_fields = ['query_execution_id']
    template_ext = ()
    ui_color = '#9BBBE0'

    @apply_defaults
    def __init__(self,
                 *,
                 query_execution_id: str,
                 aws_conn_id: str="aws_default",
                 vars_key: str = None,
                 do_var_push: bool = False,
                 sleep_time=100,
                 transform: Function=None,
                 results: Function = None,
                 **kwargs):
        super(AthenaResultOperator, self).__init__( **kwargs)        
        self.do_var_push = do_var_push
        self.aws_conn_id = aws_conn_id
        self.vars_key = vars_key
        self.results = results
        self.transform = transform
        self.sleep_time = sleep_time
        self.query_execution_id = query_execution_id

    def execute(self, context):
        results = self.hook.get_query_results(self.query_execution_id)
        if inspect.isfunction(self.transform): results = self.transform(data_set=results)
        if self.do_var_push and self.vars_key and results:
            Variable.set(self.vars_key, json.dumps(results))
        return results
  
    @cached_property
    def hook(self) -> AWSAthenaHook:
        """Create and return an AWSAthenaHook"""
        return AWSAthenaHook(self.aws_conn_id, sleep_time=self.sleep_time)