from airflow.models.baseoperator import BaseOperator
from airflow.utils.decorators import apply_defaults
from airflow.contrib.hooks.aws_athena_hook import AWSAthenaHook
from datetime import datetime, timedelta
from uuid import uuid4


class checkDataInAthenaOperator(BaseOperator):

    @apply_defaults
    def __init__(
            self,
            table_name: str,
            is_partition_time_column: bool,
            time_column_is_string: bool,
            minimum_row_count: int = 1,
            time_column_name_if_not_partition_and_year_month_day: str = None,
            timedelta_days: int = 0,
            aws_conn_id=None,
            partition_time_column_name: str  = 'partition_time',
            execution_date=None,
            *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.table_name = table_name
        self.is_partition_time_column = is_partition_time_column
        self.minimum_row_count = minimum_row_count
        self.time_column_is_string = time_column_is_string
        self.time_column_name_if_not_partition_and_year_month_day = time_column_name_if_not_partition_and_year_month_day
        self.timedelta_days = timedelta_days
        self.aws_conn_id = aws_conn_id
        self.partition_time_column_name = partition_time_column_name
        self.hook = self.get_hook()
        self.execution_date = execution_date

    def execute(self, context):
        self.hook = self.get_hook()
        execution_date = self.execution_date if self.execution_date else context.get("ds")
        where_filter = self.generate_where(execution_date)
        query = f"SELECT COUNT(*) FROM {self.table_name} {where_filter} HAVING COUNT(*) >= {self.minimum_row_count}"
        self.log.info(query)
        query_execution_id = self.hook.run_query(query,
                                                 query_context={
                                                     'Database': 'bigabid'
                                                 },
                                                 result_configuration={
                                                     'OutputLocation': 's3://bigabid-tempfiles/airflow-logs/athena/'
                                                 },
                                                 client_request_token=str(uuid4()))

        query_status = self.hook.poll_query_status(query_execution_id, max_tries=5)

        if query_status in AWSAthenaHook.FAILURE_STATES:
            error_message = self.hook.get_state_change_reason(query_execution_id)
            raise Exception(f'Final state of Athena job is {query_status}, query_execution_id is {query_execution_id}. Error: {error_message}')
        elif not query_status or query_status in AWSAthenaHook.INTERMEDIATE_STATES:
            raise Exception(
                f'Final state of Athena job is {query_status}. '
                f'Max tries of poll status exceeded, query_execution_id is {query_execution_id}.')

        results_dict = self.hook.get_query_results(query_execution_id)

        if len(results_dict['ResultSet']['Rows']) > 1:

            return query_execution_id
        else:
            raise ValueError(f"Error in checking yesterdays data for {self.table_name}")

    def get_hook(self):
        return AWSAthenaHook(self.aws_conn_id)

    def generate_where(self, execution_date):
        date = datetime.fromisoformat(execution_date) - timedelta(days=self.timedelta_days)
        year = date.strftime("%Y")
        month = date.strftime("%m")
        day = date.strftime("%d")

        if self.is_partition_time_column:
            return f"WHERE {self.partition_time_column_name} = '{execution_date}'" if self.time_column_is_string else f"WHERE {self.partition_time_column_name} = date('{execution_date}')"

        elif self.time_column_name_if_not_partition_and_year_month_day:
            return f"WHERE {self.time_column_name_if_not_partition_and_year_month_day} = '{execution_date}'" if self.time_column_is_string else f"WHERE {self.time_column_name_if_not_partition_and_year_month_day} = date('{execution_date}')"

        return f"WHERE year = '{year}' and month = '{month}' and day = '{day}'" if self.time_column_is_string else f"WHERE year = {year} and month = {month} and day = {day}"
