from typing import Optional
from airflow.utils.decorators import apply_defaults
from airflow.contrib.operators.emr_add_steps_operator import EmrAddStepsOperator
from airflow.exceptions import AirflowException
from airflow_bigabid_core.helpers.constants import DEFAULT_EMR_STEPS
from datetime import date, timedelta
from dateutil.relativedelta import *
import calendar


class CustomEmrAddStepsOperator(EmrAddStepsOperator):
    valid_modes = [
        'date_range', 'single_date', "pm_daily_data_preparation",
        "predictive_matching_rt", "predictive_matching_has_slots", 
        "predictive_matching_temporal", "user_price"]


    @apply_defaults
    def __init__(self,                 
                 step_args: list,
                 default_emr_steps: list = DEFAULT_EMR_STEPS,
                 job_flow_id: Optional[str] = None,
                 job_flow_name: Optional[str] = None,
                 data_xcom_task_id: str = None,
                 data_xcom_key: str = None,
                 job_flow_from_xcom = {},
                 mode='date_range',
                 timedelta_days: int = 0,
                 execution_date=None,
                 end_date_plus_days: int = 1,
                 *args,
                 **kwargs):
        if not job_flow_id and job_flow_from_xcom and job_flow_from_xcom.get('task_ids'):
            job_flow_id = 'from_xcom'
        kwargs['job_flow_id']=job_flow_id   
        kwargs['job_flow_name']=job_flow_name        
        super(CustomEmrAddStepsOperator, self).__init__(*args, **kwargs)
        self.default_emr_steps=default_emr_steps
        self.step_args = step_args
        self.data_xcom_task_id = data_xcom_task_id
        self.data_xcom_key = data_xcom_key
        self.mode = mode
        self.timedelta_days = timedelta_days
        self.job_flow_from_xcom = job_flow_from_xcom
        self.execution_date = execution_date
        self.end_date_plus_days = end_date_plus_days
        if self.mode not in self.valid_modes:
            raise AirflowException(
                "The mode must be one of {valid_modes},"
                "'{d}.{t}'; received '{m}'."
                .format(valid_modes=self.valid_modes,
                        d=self.dag.dag_id if self.dag else "",
                        t=self.task_id, m=self.mode))

    def pre_execute(self, context):
        task_ids = self.job_flow_from_xcom.get('task_ids')
        key = self.job_flow_from_xcom.get('key')
        if self.job_flow_id=='from_xcom' and task_ids: 
            req = dict(
                task_ids=task_ids
            )
            if key: req['key'] = key 
            self.job_flow_id=context['ti'].xcom_pull(**req) 
            
    def execute(self, context):
        data = ''
        if self.data_xcom_task_id:
            if self.data_xcom_key:
                data = context['ti'].xcom_pull(
                    task_ids=self.data_xcom_task_id, key=self.data_xcom_key)
            else:
                data = context['ti'].xcom_pull(task_ids=self.data_xcom_task_id)

        execution_date = self.execution_date if self.execution_date else context.get("ds")
        self.steps = self.generate_steps(execution_date, data)
        return super().execute(context)

    def generate_steps(self, execution_date, additional_args):
        start_date = execution_date
        if self.mode == 'date_range':
            end_date = (date.fromisoformat(execution_date) +
                        timedelta(days=self.end_date_plus_days)).isoformat()
            start_date = (date.fromisoformat(execution_date) -
                          timedelta(days=self.timedelta_days)).isoformat()
            self.step_args.extend(['--start', start_date, '--end', end_date])
        elif self.mode == 'single_date':
            self.step_args.extend(['--date', start_date])

        if additional_args:
            self.step_args.extend(additional_args.split())

        if self.mode == 'predictive_matching_rt':
            run_steps = self.default_emr_steps.copy()

            run_steps.append(
                {
                    "HadoopJarStep": {
                        "Args": ["sudo", "bash", "/home/hadoop/predictive_matching_spark/prepare_egg_for_pm.sh"],
                        "Jar": "command-runner.jar",
                        "Properties": []
                    },
                    "ActionOnFailure": "CANCEL_AND_WAIT",
                    "Name": "create .egg file"
                }
            )

            end_date = (date.fromisoformat(execution_date)).isoformat()
            start_date = (date.fromisoformat(execution_date) +
                          timedelta(days=-self.timedelta_days)).isoformat()
            target_day = (date.fromisoformat(execution_date) +
                              relativedelta(days=1)).isoformat()
            self.step_args.extend(['--gallery-start-date', start_date,
                                  '--gallery-end-date', end_date, "--target-day", target_day])
        elif self.mode == 'predictive_matching_has_slots':
            run_steps = self.default_emr_steps.copy()

            run_steps.append(
                {
                    "HadoopJarStep": {
                        "Args": ["sudo", "bash", "/home/hadoop/predictive_matching_spark/prepare_egg_for_pm.sh"],
                        "Jar": "command-runner.jar",
                        "Properties": []
                    },
                    "ActionOnFailure": "CANCEL_AND_WAIT",
                    "Name": "create .egg file"
                }
            )

            end_date = (date.fromisoformat(execution_date) +
                          relativedelta(days=6)).isoformat()
            start_date = (date.fromisoformat(end_date) +
                          timedelta(days=-self.timedelta_days)).isoformat()
            target_day = (date.fromisoformat(end_date) +
                          relativedelta(days=1)).isoformat()
            self.step_args.extend([
                '--gallery-start-date', start_date, '--gallery-end-date', end_date,
                "--target-day", target_day])
            
        elif self.mode == 'predictive_matching_temporal':
            run_steps = self.default_emr_steps.copy()

            run_steps.append(
                {
                    "HadoopJarStep": {
                        "Args": ["sudo", "bash", "/home/hadoop/predictive_matching_spark/prepare_egg_for_pm.sh"],
                        "Jar": "command-runner.jar",
                        "Properties": []
                    },
                    "ActionOnFailure": "CANCEL_AND_WAIT",
                    "Name": "create .egg file"
                }
            )

            end_date = (date.fromisoformat(execution_date)).isoformat()
            start_date = (date.fromisoformat(end_date) +
                          relativedelta(months=-2)).isoformat()
            target_day = (date.fromisoformat(execution_date) +
                          relativedelta(days=1)).isoformat()
            self.step_args.extend([
                '--gallery-start-date', start_date, '--gallery-end-date', end_date,
                "--target-day", target_day])

        elif self.mode == 'user_price':
            run_steps = self.default_emr_steps.copy()
            
        elif self.mode == "pm_daily_data_preparation":
            
            run_steps = self.default_emr_steps.copy()
            
            today = date.today().isoformat()
            next_day_date = (date.fromisoformat(execution_date) + relativedelta(days=+1)).isoformat()
            if next_day_date == today:
                
                if calendar.day_name[date.today().weekday()] == "Monday": # if today is Monday, than run CRM with all apps (slots and rt apps)
                    self.step_args.extend([
                    '--execution-date', execution_date, "--mode", "rt_slots"])
                else: # CRM run with rt APPs
                    self.step_args.extend([
                    '--execution-date', execution_date, "--mode", "rt"])
            else: # then backfill
                self.step_args.extend([
                '--execution-date', execution_date, "--skip-crm"])
                
            run_steps.append(
                {
                    "HadoopJarStep": {
                        "Args": ["sudo", "bash", "/home/hadoop/pm_daily_data_preparation/prepare_egg_file.sh"],
                        "Jar": "command-runner.jar",
                        "Properties": []
                    },
                    "ActionOnFailure": "CANCEL_AND_WAIT",
                    "Name": "create .egg file"
                }
            )
        else:
            run_steps = self.default_emr_steps.copy()

        run_steps.append({
            "HadoopJarStep": {
                "Args": self.step_args,
                "Jar": "command-runner.jar",
                "Properties": []
            },
            "ActionOnFailure": "CANCEL_AND_WAIT",
            "Name": "SparkApp"
        })
        return run_steps
