from typing import List, Dict
from airflow.utils.decorators import apply_defaults
from airflow.contrib.operators.emr_create_job_flow_operator import EmrCreateJobFlowOperator


class CustomEmrOperator(EmrCreateJobFlowOperator):
    @apply_defaults
    def __init__(
            self,
            cluster_name: str,
            bootstrap_script_path: str,
            emr_version: str,
            master_instance_types_dict: dict,
            master_units: int,
            core_instance_types_dict: dict,
            core_units: int,
            task_instance_types_dict: dict,
            task_units: int,
            spark_executor_memory: str,
            spark_executor_cores: str,
            master_ebs_storage_per_type: dict = None,
            core_ebs_storage_per_type: dict = None,
            task_ebs_storage_per_type: dict = None,
            master_target_capacity_type: str = 'TargetOnDemandCapacity',
            core_target_capacity_type: str = 'TargetSpotCapacity',
            task_target_capacity_type: str = 'TargetSpotCapacity',
            spark_properties: dict = {},
            yarn_properties: dict = {},
            managed_scaling_policy: dict = None,
            hive_site_properties: dict = {},
            spark_hive_site_properties: dict = {},
            bootstrap_git_project_name: str = None,
            bootstrap_git_project_branch: str = 'master',
            master_ebs_storage_in_gb: int = None,
            core_ebs_storage_in_gb: int = None,
            task_ebs_storage_in_gb: int = None,
            keep_job_flow_alive_when_no_steps: bool = False,
            cluster_idle_timeout_seconds: int = None,
            spark_jars_packages: List=[],
            tags: dict = None,
            retry_number: str = '{{ ti.try_number }}',
            *args, **kwargs):
        self.managed_scaling_policy = managed_scaling_policy
        self.spark_properties = spark_properties
        self.yarn_properties = yarn_properties
        self.hive_site_properties = hive_site_properties
        self.spark_hive_site_properties = spark_hive_site_properties
        self.cluster_name = cluster_name
        self.tags = [{"Value": self.cluster_name, "Key": "Name"}]
        if retry_number:
                tag={"Key": "retry", "Value":f"Retry_attempt: {retry_number}"}
                self.tags.append(tag)
        if tags:
            assert all(tag.keys() == set(["Value", "Key"]) for tag in tags)
            self.tags += tags
        else:
            self.tags += [{"Key": "component", "Value": "ETL"},
                          {"Key": "team", "Value": "DE"},
                          {"Key": "env", "Value": "production"}]
        self.keep_job_flow_alive_when_no_steps = keep_job_flow_alive_when_no_steps
        self.cluster_idle_timeout_seconds=cluster_idle_timeout_seconds
        self.bootstrap_script_path = bootstrap_script_path
        self.bootstrap_git_project_name = bootstrap_git_project_name
        self.bootstrap_git_project_branch = bootstrap_git_project_branch
        self.emr_version = emr_version
        self.spark_executor_memory = spark_executor_memory
        self.spark_executor_cores = spark_executor_cores
        self.spark_jars_packages = spark_jars_packages
        self.fleet_type_to_variables = {
            "MASTER": {
                "ebs": master_ebs_storage_in_gb,
                "instance_types": master_instance_types_dict,
                "instance_units": master_units,
                "ebs_storage_per_type": master_ebs_storage_per_type,
                "target_capacity" : master_target_capacity_type
            },
            "CORE": {
                "ebs": core_ebs_storage_in_gb,
                "instance_types": core_instance_types_dict,
                "instance_units": core_units,
                "ebs_storage_per_type": core_ebs_storage_per_type,
                "target_capacity" : core_target_capacity_type

            },
            "TASK": {
                "ebs": task_ebs_storage_in_gb,
                "instance_types": task_instance_types_dict,
                "instance_units": task_units,
                "ebs_storage_per_type": task_ebs_storage_per_type,
                "target_capacity" : task_target_capacity_type
            }
        }
        super(CustomEmrOperator, self).__init__(job_flow_overrides=self.generate_emr_configuration(), *args, **kwargs)

    def generate_script_bootstrap_action(self):
        script_bootstrap = {
            "Path": self.bootstrap_script_path
        }

        if self.bootstrap_git_project_name:
            script_bootstrap['Args'] = [self.bootstrap_git_project_name, self.bootstrap_git_project_branch]

        return script_bootstrap

    def generate_single_instance_fleet(self, instance_fleet_type: str):
        

        fleet_variables = self.fleet_type_to_variables[instance_fleet_type]

        instance_fleet = {
                "InstanceFleetType": instance_fleet_type,
                "InstanceTypeConfigs": generate_instance_type_configs(fleet_variables['instance_types'], fleet_variables['ebs_storage_per_type']),
                fleet_variables['target_capacity']: fleet_variables['instance_units'],
                "Name": instance_fleet_type
            }
            
        if fleet_variables['target_capacity'] == 'TargetSpotCapacity':
            instance_fleet["LaunchSpecifications"] = {
                "SpotSpecification": {
                    "AllocationStrategy": "capacity-optimized",
                    "TimeoutAction": "TERMINATE_CLUSTER",
                    "TimeoutDurationMinutes": 60,
                }
            }
        elif fleet_variables['target_capacity'] == 'TargetOnDemandCapacity':
            instance_fleet["LaunchSpecifications"] = {
                "OnDemandSpecification": {
                    "AllocationStrategy": "lowest-price"
                }
            }
            
        if not fleet_variables['ebs_storage_per_type'] and fleet_variables['ebs']:
            for instance_type_config in instance_fleet["InstanceTypeConfigs"]:
                instance_type_config['EbsConfiguration'] = set_ebs_conf(
                    fleet_variables['ebs'])

        return instance_fleet

    def generate_instance_fleets(self):
        instance_fleets_list = list()
        instance_fleets_list.append(self.generate_single_instance_fleet('MASTER'))
        instance_fleets_list.append(self.generate_single_instance_fleet('CORE'))
        instance_fleets_list.append(self.generate_single_instance_fleet('TASK'))
        return instance_fleets_list

    def generate_emr_configuration(self):


        spark_default_properties_dict = {
            "spark.executor.memory": self.spark_executor_memory,
            "spark.executor.cores": self.spark_executor_cores,
            "spark.driver.maxResultSize": "0",
            "spark.jars.packages": ','.join(self.spark_jars_packages),
            "spark.jars.repositories": "https://repos.spark-packages.org/",
            "spark.hadoop.fs.s3.maxRetries":"30",
            "spark.hadoop.fs.s3.consistent.retryPolicyType":"exponential"
        }

        if self.spark_properties:
            spark_default_properties_dict.update(self.spark_properties)

        #Default settings to enable hive
        hive_site_properties_final =  {"hive.metastore.client.factory.class":"com.amazonaws.glue.catalog.metastore.AWSGlueDataCatalogHiveClientFactory"}
        spark_hive_site_properties_final =  {"hive.metastore.client.factory.class":"com.amazonaws.glue.catalog.metastore.AWSGlueDataCatalogHiveClientFactory"}

        #update default with user properties
        hive_site_properties_final.update(self.hive_site_properties)
        spark_hive_site_properties_final.update(self.spark_hive_site_properties)

        emr_conf = {
            "Instances": {
                "KeepJobFlowAliveWhenNoSteps": self.keep_job_flow_alive_when_no_steps,
                "EmrManagedMasterSecurityGroup": "sg-526e8e34",
                "Ec2KeyName": "ds-team",
                "Ec2SubnetIds": [ "subnet-4b711e3c", "subnet-7cee8c70", "subnet-874b34e2", "subnet-ae62eff7", "subnet-c73386ec", "subnet-e24c11d8" ],
                "EmrManagedSlaveSecurityGroup": "sg-536e8e35",
                "TerminationProtected": False,
                "InstanceFleets": self.generate_instance_fleets()
            },
            "Name": self.cluster_name,
            "BootstrapActions": [
                {
                    "Name": "initialize cluster",
                    "ScriptBootstrapAction": self.generate_script_bootstrap_action()
                }
            ],
            "Configurations": [
                {
                    "Classification": "spark",
                    "Configurations": [],
                    "Properties": {
                        "maximizeResourceAllocation": "true"
                    }
                },                                 
                {
                    "Classification": "yarn-site",
                    "Configurations": [],
                    "Properties": self.yarn_properties
                },
                {
                    "Classification": "spark-defaults",
                    "Configurations": [],
                    "Properties": spark_default_properties_dict
                },
                                {
                    "Classification": "hive-site",
                    "Configurations": [],
                    "Properties":  hive_site_properties_final
                } ,
                {
                    "Classification": "spark-hive-site",
                    "Configurations": [],
                    "Properties":  spark_hive_site_properties_final
                }
            ],
            "LogUri": "s3n://bigabid-ml/logs/",
            "Tags": self.tags,
            "Applications": [
                {
                    "Name": "Spark"
                }
            ],
            "EbsRootVolumeSize": 20,
            "ScaleDownBehavior": "TERMINATE_AT_TASK_COMPLETION",
            "JobFlowRole": "EMR_EC2_DefaultRole",
            "ServiceRole": "EMR_DefaultRole",
            "ReleaseLabel": self.emr_version,
            "VisibleToAllUsers": True
        }

        if self.managed_scaling_policy:
            emr_conf["ManagedScalingPolicy"] = {"ComputeLimits": self.managed_scaling_policy}

        if self.cluster_idle_timeout_seconds:
            emr_conf["AutoTerminationPolicy"] = {"IdleTimeout": self.cluster_idle_timeout_seconds}
           
        return emr_conf



def set_ebs_conf(size_in_gb: int):
    return {
        "EbsBlockDeviceConfigs": [
            {
                "VolumeSpecification": {
                    "SizeInGB": size_in_gb,
                    "VolumeType": "gp2"
                },
                "VolumesPerInstance": 1
            }
        ],
        "EbsOptimized": True
    }


def generate_instance_type_configs(instance_types, ebs_storage_per_type):
    instance_type_configs_list = list()
    for instance_type, weighted_capacity in instance_types.items():
        instance_type_configs_dict = {
            "InstanceType": instance_type,
            "BidPriceAsPercentageOfOnDemandPrice": 100,
            "WeightedCapacity": weighted_capacity,
        }
        if ebs_storage_per_type and ebs_storage_per_type[instance_type] > 0:
            instance_type_configs_dict['EbsConfiguration'] = \
                {'EbsBlockDeviceConfigs': [
                    {
                        'VolumeSpecification': {
                            'VolumeType': 'gp2',
                            'SizeInGB': ebs_storage_per_type[instance_type]
                        }
                    },
                ]}
        instance_type_configs_list.append(instance_type_configs_dict)
    return instance_type_configs_list

