# -*- coding: utf-8 -*-
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from airflow import AirflowException
from airflow.contrib.hooks.emr_hook import EmrHook
from airflow.contrib.sensors.emr_base_sensor import EmrBaseSensor
from airflow.exceptions import AirflowFailException
from airflow.models import XCom
from airflow.utils.decorators import apply_defaults
from airflow.utils.session import provide_session

XCOM_RETURN_KEY = 'emr_step_retry'

class EmrExtendedStepSensor(EmrBaseSensor):
    """
    Asks for the state of the step until it reaches a terminal state.
    If it fails the sensor errors, failing the task.

    :param job_flow_id: job_flow_id which contains the step check the state of
    :type job_flow_id: str
    :param step_id: step to check the state of
    :type step_id: str
    """

    NON_TERMINAL_STATES = ['PENDING', 'RUNNING', 'CONTINUE', 'CANCEL_PENDING']
    FAILED_STATE = ['CANCELLED', 'FAILED', 'INTERRUPTED']
    COMPLETED_STATE = 'COMPLETED'
    template_fields = ['job_flow_id', 'step_id']
    template_ext = ()

    # @apply_defaults
    def __init__(self,
                 job_flow_id,
                 step_id,
                 max_tries=2,
                 *args,
                 **kwargs):
              
        self.job_flow_id = job_flow_id
        self.step_id = step_id        
        self.max_tries=max_tries
        self.retry_count=max_tries
        self._emr_hook=None
        self.log.info('im here %s', self.retry_count)
        super(EmrExtendedStepSensor, self).__init__(*args, **kwargs)       

    @property
    def emr_hook(self):
        if not self._emr_hook:
           self._emr_hook = EmrHook(aws_conn_id=self.aws_conn_id).get_conn()
        return self._emr_hook
        
    def get_emr_response(self):    
        self.log.info('get emr response for step with step_id %s on cluster with job_flow_id %s', self.step_id, self.job_flow_id)
        return self.emr_hook.describe_step(ClusterId=self.job_flow_id, StepId=self.step_id)

    @staticmethod
    def state_from_response(response):
        return response['Step']['Status']['State']
    
    @property
    def task_id_key(self):
        return f"{self.task_id}_cp"

    @provide_session
    def xcom_delete(self, context, session=None):    
        session.query(XCom).filter(
            XCom.key == XCOM_RETURN_KEY, XCom.execution_date == context['ti'].execution_date, XCom.task_id == self.task_id_key, XCom.dag_id == self.dag_id
        ).delete()
  
    def xcom_pull(self, context):
        return super().xcom_pull(
            context=context,
            key=XCOM_RETURN_KEY, 
            task_ids=self.task_id_key, 
        )

    def xcom_push(self, context):
        XCom.set(
            key=XCOM_RETURN_KEY,
            value=self.retry_count,
            task_id=self.task_id_key,
            dag_id=self.dag_id,
            execution_date=context['ti'].execution_date
        )

    def poke(self, context):
        try:
            
            self.log.info('inside poke, retry_count %s out of max_tries %s', self.max_tries-self.retry_count, self.max_tries)
            
            retry_from_xcom = self.xcom_pull(context)
            if not retry_from_xcom:
                self.log.info(f'retry_count: {self.retry_count}')
            else:
                self.retry_count = retry_from_xcom
                self.log.info(f'retry_count from xcom: {self.retry_count}')

            response = self.get_emr_response()

            if not response['ResponseMetadata']['HTTPStatusCode'] == 200:
                self.log.info('bad HTTP response: %s', response)
                return False

            state = self.state_from_response(response)
            self.log.info('Job flow currently %s', state)
            
            if state in self.NON_TERMINAL_STATES:
                self.log.info('NON_TERMINAL_STATES')
                self.xcom_push(context)
                return False

            if state == self.COMPLETED_STATE:
                self.log.info('COMPLETED_STATE')
                return True
            
            self.log.info("state failed")
            self.log.info(f"retry_count {self.retry_count}")
            
            if self.retry_count > 0:
                prior_step_response = self.get_emr_response()    
                self.log.info('retry steps %s added to jobflow', prior_step_response['Step']['Id'])            
                if not prior_step_response['ResponseMetadata']['HTTPStatusCode'] == 200:
                    self.xcom_delete(context=context)
                    raise AirflowException('Describing prior step failed: %s' % prior_step_response)
                retry_step = self.map_retry_step(prior_step_response)
                response = self.emr_hook.add_job_flow_steps(JobFlowId=self.job_flow_id, Steps=retry_step)
                if not response['ResponseMetadata']['HTTPStatusCode'] == 200:
                    self.xcom_delete(context=context)
                    raise AirflowException('adding Retry Step Failed: %s' % response)
                else:
                    self.log.info('retry steps %s added to jobflow', response['StepIds'])
                    self.step_id = response['StepIds'][0]
                    self.retry_count-=1
                    self.xcom_push(context)
                    self.log.info("attempt number: %s", self.retry_count)
                    return False
            else:
                self.log.info('retry count %s reached max_tries %s', self.max_tries-self.retry_count, self.max_tries)
                final_message = 'EMR step failed'
                failure_message = self.failure_message_from_response(response)
                self.xcom_delete(context=context)
                if failure_message:
                    final_message += f' {failure_message}'
                    raise AirflowFailException(final_message)
        except Exception as e:
            self.log.exception('failed to poke emr step')                  
            self.xcom_delete(context=context)
            raise e

    def map_retry_step(self, response):
        steps = [{
            'Name': response['Step']['Name'],
            'ActionOnFailure': response['Step']['ActionOnFailure'],
            'HadoopJarStep': {
                'Jar': response['Step']['Config']['Jar'],
                'Args': response['Step']['Config']['Args']
            }
        }]
        return steps

    @staticmethod
    def failure_message_from_response(response):
        fail_details = response['Step']['Status'].get('FailureDetails')
        if fail_details:
            return 'for reason {} with message {} and log file {}'.format(fail_details.get('Reason'),
                                                                          fail_details.get('Message'),
                                                                          fail_details.get('LogFile'))
        return None
