# -*- coding: utf-8 -*-
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
import logging
from typing import Any, Dict
from airflow.contrib.hooks.emr_hook import EmrHook
from airflow.contrib.sensors.emr_base_sensor import EmrBaseSensor
from airflow.utils.decorators import apply_defaults


class EmrJobInstanceFleetSensor(EmrBaseSensor):
    """
    Asks for the state of the JobFlow until it reaches a terminal state.
    If it fails the sensor errors, failing the task.
    :param job_flow_id: job_flow_id to check the state of
    :type job_flow_id: str
    """
    template_fields = ['job_flow_id']
    template_ext = ()

    @apply_defaults
    def __init__(self,
                 job_flow_id=None,
                 job_flow_from_xcom={},
                 fleet_type='TASK',
                 required_running_units_percent=100,
                 *args,
                 **kwargs):
        super(EmrJobInstanceFleetSensor, self).__init__(*args, **kwargs)
        self.job_flow_id = job_flow_id
        self.job_flow_from_xcom = job_flow_from_xcom
        self.target_states = ['RUNNING'] #State on which we can move to the next task
        self.failed_states = ['TERMINATING', 'TERMINATED']
        self.fleet_type = fleet_type
        self.required_running_units_percent = required_running_units_percent
    
    def get_running_units_percent(self, client, if_data):
        cluster_id = self.job_flow_id or self.job_flow_from_xcom
        InstanceTypeSpecifications = if_data['InstanceTypeSpecifications']
        target_spot_capacity = int(if_data['TargetSpotCapacity'])
        instance_type_to_units_map = {spec['InstanceType']: int(spec['WeightedCapacity']) for spec in InstanceTypeSpecifications}

        task_instances_resp = client.list_instances(
            ClusterId=cluster_id,
            InstanceGroupTypes=[self.fleet_type]
        )
        task_instances = task_instances_resp['Instances']

        running_units = 0
        for instance in task_instances:
            if instance['Status']['State'] == 'RUNNING':
                running_units += instance_type_to_units_map[instance['InstanceType']]
        
        running_units_percent = (running_units / target_spot_capacity) * 100
        return running_units_percent


    def get_emr_response(self):
        emr = EmrHook(aws_conn_id=self.aws_conn_id).get_conn()
        self.log.info('Poking cluster %s', self.job_flow_id)
        response = emr.list_instance_fleets(ClusterId=self.job_flow_id)

        fleet_data = [f for f in response['InstanceFleets'] if f['Name'] == self.fleet_type][0]
        fleet_data['ResponseMetadata'] = response['ResponseMetadata']

        if self.required_running_units_percent == 100:
            return fleet_data
        
        running_units_percent = self.get_running_units_percent(emr, fleet_data)
        logging.info(f"Instance fleet running instances percent: {running_units_percent}% , required: {self.required_running_units_percent}%")
        if running_units_percent >= self.required_running_units_percent:
            fleet_data['Status']['State'] = 'RUNNING'
            logging.info(f"running_units_percent requirement satisfied")
            logging.info(fleet_data['Status']['State'])
            return fleet_data
        else:
            return fleet_data


    @staticmethod
    def state_from_response(response: Dict[str, Any]) -> str:
        """
        Get state from response dictionary.

        :param response: response from AWS API
        :type response: dict[str, Any]
        :return: state
        :rtype: str
        """
    
        logging.info(f"TASK instance fleet state: {response['Status']['State']}")
        return response['Status']['State']

    def pre_execute(self, context):
        if self.job_flow_from_xcom:
            job_flow_tasks_id = self.job_flow_from_xcom.get('task_ids')
            job_flow_key = self.job_flow_from_xcom.get('key')
            if not self.job_flow_id and job_flow_tasks_id: 
                req = dict(
                    task_ids=job_flow_tasks_id
                )
                if job_flow_key: req['key'] = job_flow_key 
                self.job_flow_id=context['ti'].xcom_pull(**req) 
                self.log.info('job_flow_id: %s', self.job_flow_id)
            
    @staticmethod
    def failure_message_from_response(response):
        task_if = [f for f in response['InstanceFleets'] if f['Name'] == 'TASK'][0]
        state_change_reason = task_if['Status'].get('StateChangeReason')
        if state_change_reason:
            return 'for code: {} with message {}'.format(state_change_reason.get('Code', 'No code'),
                                                         state_change_reason.get('Message', 'Unknown'))
        return None