# -*- coding: utf-8 -*-
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from airflow.contrib.hooks.emr_hook import EmrHook
from airflow.contrib.sensors.emr_step_sensor import EmrStepSensor


class EmrStepFlowSensor(EmrStepSensor):
    """
    Asks for the state of the step until it reaches a terminal state.
    If it fails the sensor errors, failing the task.

    :param job_flow_id: job_flow_id which contains the step check the state of
    :type job_flow_id: str
    :param step_id: step to check the state of
    :type step_id: str
    """

    NON_TERMINAL_STATES = ['PENDING', 'RUNNING', 'CONTINUE', 'CANCEL_PENDING']

    FAILED_STATE = ['CANCELLED', 'FAILED', 'INTERRUPTED']

    template_fields = ['job_flow_id', 'step_id']

    template_ext = ()


    def __init__(self,
                 job_flow_from_xcom={},
                 step_flow_from_xcom={},
                 job_flow_id = None,
                 step_id = None,
                 *args,
                 **kwargs):
        kwargs['job_flow_id']=job_flow_id   
        kwargs['step_id']=step_id   
        super(EmrStepFlowSensor, self).__init__(*args, **kwargs)
        self.job_flow_id = job_flow_id
        self.step_id = step_id
        self.job_flow_from_xcom = job_flow_from_xcom
        self.step_flow_from_xcom = step_flow_from_xcom

    def get_emr_response(self):
        emr = EmrHook(aws_conn_id=self.aws_conn_id).get_conn()

        self.log.info('Poking step %s on cluster %s', self.step_id, self.job_flow_id)
        return emr.describe_step(ClusterId=self.job_flow_id, StepId=self.step_id)

    def pre_execute(self, context):
        if self.job_flow_from_xcom:
            job_flow_task_ids = self.job_flow_from_xcom.get('task_ids')
            job_flow_key = self.job_flow_from_xcom.get('key')
            if not self.job_flow_id and job_flow_task_ids: 
                req = dict(
                    task_ids=job_flow_task_ids
                )
                if job_flow_key: req['key'] = job_flow_key 
                self.job_flow_id=context['ti'].xcom_pull(**req) 

        if self.step_flow_from_xcom:
            step_flow_task_ids = self.step_flow_from_xcom.get('task_ids')
            step_flow_key = self.step_flow_from_xcom.get('key')
            step_index = self.step_flow_from_xcom.get('index')
            if not self.step_id and step_flow_task_ids: 
                req = dict(
                    task_ids=step_flow_task_ids
                )
                if step_flow_key: req['key'] = step_flow_key
                self.step_id=context['ti'].xcom_pull(**req)[step_index]